From 4db7adfe27ac216cbeb2b2f34101dcd537a850a7 Mon Sep 17 00:00:00 2001
From: Peter Geoghegan <pg@bowt.ie>
Date: Wed, 17 Apr 2019 20:17:19 -0700
Subject: [PATCH] Prevent O(N^2) unique index insertion edge case.

Commit dd299df8 made nbtree treat heap TID as a tiebreaker column,
establishing the principle that there is only one correct location (leaf
page and offset number) for every newly inserted index tuple, no matter
what.  Insertions into non-unique indexes can proceed as if heap TID
(scan key's scantid) is just another user-attribute value, but
insertions into unique indexes must be more careful.  The TID value in
scantid must initially be omitted to reliably visit every leaf page that
duplicates could be on, which is needed by unique checking.  The scantid
is filled in after unique checking finishes successfully, which
occasionally forces _bt_findinsertloc() to step right to find the leaf
page that the new tuple must be inserted on.

Unique checking typically has very little overhead, since in practice it
is hardly ever necessary to actually step to the right within
_bt_findinsertloc().  There seems to be no realistic way that an
accretion of _bt_check_unique()-wise duplicates of a single value (or
set of values) can slow down insertions of the same value over time.
Even if there are duplicates of a single value that span multiple leaf
pages, it's inherently necessary to visit those leaf pages anyway, to
check visibility information from the heap.  Visiting the same leaf
pages a second time (when we get to _bt_findinsertloc()) is not
expensive in comparison.

However, there is an important distinction that commit dd299df8
overlooked: the distinction between duplicates that have no NULL values
(tuples that can be involved in unique violation errors), and tuples
with NULL values.  A huge number of duplicates with NULL values are
common with some workloads, which caused problematic O(N^2) behavior
within _bt_findinsertloc()'s heapkeyspace path for later insertions of
tuples of the same value.  In extreme cases, _bt_findinsertloc() could
end up grovelling through every leaf page in the index.

To fix, treat insertions of tuples with NULLs into a unique index as if
they were insertions into a non-unique index:

* Never set scantid to NULL before calling _bt_search() to descend the
  tree when the new tuple contains a NULL value.

* Never perform unique checking.

This allows us to remove code that was used by _bt_check_unique() to
handle NULL values.

Discussion: https://postgr.es/m/CAH2-Wzm08nr+JPx4jMOa9CGqxWYDQ-_D4wtPBiKghXAUiUy-nQ@mail.gmail.com
---
 src/backend/access/nbtree/nbtinsert.c | 106 +++++++++-----------------
 1 file changed, 34 insertions(+), 72 deletions(-)

diff --git a/src/backend/access/nbtree/nbtinsert.c b/src/backend/access/nbtree/nbtinsert.c
index 4fda0efe9a..5b85ede480 100644
--- a/src/backend/access/nbtree/nbtinsert.c
+++ b/src/backend/access/nbtree/nbtinsert.c
@@ -55,8 +55,6 @@ static void _bt_insert_parent(Relation rel, Buffer buf, Buffer rbuf,
 				  BTStack stack, bool is_root, bool is_only);
 static bool _bt_pgaddtup(Page page, Size itemsize, IndexTuple itup,
 			 OffsetNumber itup_off);
-static bool _bt_isequal(TupleDesc itupdesc, BTScanInsert itup_key,
-			Page page, OffsetNumber offnum);
 static void _bt_vacuum_one_page(Relation rel, Buffer buffer, Relation heapRel);
 
 /*
@@ -91,9 +89,28 @@ _bt_doinsert(Relation rel, IndexTuple itup,
 
 	/* we need an insertion scan key to do our search, so build one */
 	itup_key = _bt_mkscankey(rel, itup);
-	/* No scantid until uniqueness established in checkingunique case */
-	if (checkingunique && itup_key->heapkeyspace)
-		itup_key->scantid = NULL;
+
+	if (checkingunique)
+	{
+		/* No scantid until uniqueness established in checkingunique case */
+		if (!IndexTupleHasNulls(itup))
+			itup_key->scantid = NULL;
+		else
+		{
+			/*
+			 * Bypass _bt_check_unique() when new tuple contains NULL value,
+			 * and assume that it is "unique" for caller's purposes.  This
+			 * optimization helps insertions into a unique index with many
+			 * duplicate tuples with NULLs.  Descending directly to the page
+			 * that we'll ultimately insert on (by keeping scantid set)
+			 * prevents O(N^2) behavior within the _bt_findinsertloc()
+			 * heapkeyspace path.
+			 */
+			Assert(checkUnique != UNIQUE_CHECK_EXISTING);
+			is_unique = true;
+			checkingunique = false;
+		}
+	}
 
 	/*
 	 * Fill in the BTInsertState working area, to track the current page and
@@ -203,8 +220,8 @@ top:
 	buf = InvalidBuffer;		/* insertstate.buf now owns the buffer */
 
 	/*
-	 * If we're not allowing duplicates, make sure the key isn't already in
-	 * the index.
+	 * If we're not allowing duplicates, and new tuple has no NULLs, make sure
+	 * the key isn't already in the index.
 	 *
 	 * NOTE: obviously, _bt_check_unique can only detect keys that are already
 	 * in the index; so it cannot defend against concurrent insertions of the
@@ -300,6 +317,9 @@ top:
 /*
  *	_bt_check_unique() -- Check for violation of unique index constraint
  *
+ * Caller should not pass new tuple that contains any NULL values.  We're
+ * not prepared for that case, and a unique violation isn't possible anyway.
+ *
  * Returns InvalidTransactionId if there is no conflict, else an xact ID
  * we must wait for to see if it commits a conflicting tuple.   If an actual
  * conflict is detected, no return --- just ereport().  If an xact ID is
@@ -321,7 +341,6 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 				 IndexUniqueCheck checkUnique, bool *is_unique,
 				 uint32 *speculativeToken)
 {
-	TupleDesc	itupdesc = RelationGetDescr(rel);
 	IndexTuple	itup = insertstate->itup;
 	BTScanInsert itup_key = insertstate->itup_key;
 	SnapshotData SnapshotDirty;
@@ -354,6 +373,7 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 	 * Scan over all equal tuples, looking for live conflicts.
 	 */
 	Assert(!insertstate->bounds_valid || insertstate->low == offset);
+	Assert(!IndexTupleHasNulls(itup));
 	Assert(itup_key->scantid == NULL);
 	for (;;)
 	{
@@ -375,8 +395,8 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 			 * original page, which may indicate that we need to examine a
 			 * second or subsequent page.
 			 *
-			 * Note that this optimization avoids calling _bt_isequal()
-			 * entirely when there are no duplicates, as long as the offset
+			 * Note that this optimization avoids calling _bt_compare() within
+			 * this loop when there are no duplicates, as long as the offset
 			 * where the key will go is not at the end of the page.
 			 */
 			if (nbuf == InvalidBuffer && offset == insertstate->stricthigh)
@@ -384,7 +404,7 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 				Assert(insertstate->bounds_valid);
 				Assert(insertstate->low >= P_FIRSTDATAKEY(opaque));
 				Assert(insertstate->low <= insertstate->stricthigh);
-				Assert(!_bt_isequal(itupdesc, itup_key, page, offset));
+				Assert(_bt_compare(rel, itup_key, page, offset) < 0);
 				break;
 			}
 
@@ -394,9 +414,9 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 			 * We can skip items that are marked killed.
 			 *
 			 * In the presence of heavy update activity an index may contain
-			 * many killed items with the same key; running _bt_isequal() on
+			 * many killed items with the same key; running _bt_compare() on
 			 * each killed item gets expensive.  Just advance over killed
-			 * items as quickly as we can.  We only apply _bt_isequal() when
+			 * items as quickly as we can.  We only apply _bt_compare() when
 			 * we get to a non-killed item.  Even those comparisons could be
 			 * avoided (in the common case where there is only one page to
 			 * visit) by reusing bounds, but just skipping dead items is fast
@@ -407,13 +427,7 @@ _bt_check_unique(Relation rel, BTInsertState insertstate, Relation heapRel,
 				ItemPointerData htid;
 				bool		all_dead;
 
-				/*
-				 * _bt_compare returns 0 for (1,NULL) and (1,NULL) - this's
-				 * how we handling NULLs - and so we must not use _bt_compare
-				 * in real comparison, but only for ordering/finding items on
-				 * pages. - vadim 03/24/97
-				 */
-				if (!_bt_isequal(itupdesc, itup_key, page, offset))
+				if (_bt_compare(rel, itup_key, page, offset) != 0)
 					break;		/* we're past all the equal tuples */
 
 				/* okay, we gotta fetch the heap tuple ... */
@@ -2184,58 +2198,6 @@ _bt_pgaddtup(Page page,
 	return true;
 }
 
-/*
- * _bt_isequal - used in _bt_doinsert in check for duplicates.
- *
- * This is very similar to _bt_compare, except for NULL and negative infinity
- * handling.  Rule is simple: NOT_NULL not equal NULL, NULL not equal NULL too.
- */
-static bool
-_bt_isequal(TupleDesc itupdesc, BTScanInsert itup_key, Page page,
-			OffsetNumber offnum)
-{
-	IndexTuple	itup;
-	ScanKey		scankey;
-	int			i;
-
-	/* Better be comparing to a non-pivot item */
-	Assert(P_ISLEAF((BTPageOpaque) PageGetSpecialPointer(page)));
-	Assert(offnum >= P_FIRSTDATAKEY((BTPageOpaque) PageGetSpecialPointer(page)));
-	Assert(itup_key->scantid == NULL);
-
-	scankey = itup_key->scankeys;
-	itup = (IndexTuple) PageGetItem(page, PageGetItemId(page, offnum));
-
-	for (i = 1; i <= itup_key->keysz; i++)
-	{
-		AttrNumber	attno;
-		Datum		datum;
-		bool		isNull;
-		int32		result;
-
-		attno = scankey->sk_attno;
-		Assert(attno == i);
-		datum = index_getattr(itup, attno, itupdesc, &isNull);
-
-		/* NULLs are never equal to anything */
-		if (isNull || (scankey->sk_flags & SK_ISNULL))
-			return false;
-
-		result = DatumGetInt32(FunctionCall2Coll(&scankey->sk_func,
-												 scankey->sk_collation,
-												 datum,
-												 scankey->sk_argument));
-
-		if (result != 0)
-			return false;
-
-		scankey++;
-	}
-
-	/* if we get here, the keys are equal */
-	return true;
-}
-
 /*
  * _bt_vacuum_one_page - vacuum just one index page.
  *
-- 
2.17.1

